unit uCalibrIn;

interface

uses IniFiles, uCommon;

type
  TCalibrIn = class(TSignalProcessing)
  private
    FChannels: array of record
      AttIdx: Integer;
      AttData: array of record
        OffSet: Extended;
        AttVal: Extended;
      end;
    end;
    FRefChannel: Integer;
    FMonitorFunction: TDataFunction;
    procedure SetChannels(aValue: Integer);
    function GetChannels: Integer;
    procedure SetAttNmb(aValue: Integer);
    function GetAttNmb: Integer;
    procedure SetAtt(aChannel: Integer; aValue: Integer);
    function GetAtt(aChannel: Integer): Integer;
  public
    constructor Create;
    function Input(const aData: TDataMatrix): Integer; override;
    function InputOffSetCalibration(const aData: TDataMatrix): Integer;
    function InputAmplitudeCalibration(const aData: TDataMatrix): Integer;
    procedure SaveIni(aFile: TIniFile);
    procedure LoadIni(aFile: TIniFile);
    property Channels: Integer read GetChannels write SetChannels;
    property AttNumber: Integer read GetAttNmb write SetAttNmb;
    property Attenuator[aChannel: Integer]: Integer read GetAtt write SetAtt;
    property ReferenceChannel: Integer read FRefChannel write FRefChannel;
    property MonitorFunction: TDataFunction write FMonitorFunction;
  end;

implementation

uses Math, Sysutils;
const sCalibr = 'Calibration';

constructor TCalibrIn.Create;
begin
  inherited;
  Channels:= 2;
  FRefChannel:= 0;
end;

procedure TCalibrIn.SetChannels(aValue: Integer);
var i, j, aOldValue: Integer;
begin
  aOldValue:= Length(FChannels);
  SetLength(FChannels, aValue);
  if aOldValue >= aValue then
    Exit;
  if aOldValue = 0 then for i:= 0 to aValue - 1 do begin
    FChannels[i].AttIdx:= 0;
    SetLength(FChannels[i].AttData, 5);
    for j:= 0 to High(FChannels[i].AttData) do begin
      FChannels[i].AttData[j].OffSet:= 0;
      FChannels[i].AttData[j].AttVal:= 1 shl j;
    end;
  end else if aOldValue < aValue then for i:= aOldValue to aValue - 1 do begin
    FChannels[i].AttIdx:= FChannels[aOldValue - 1].AttIdx;
    SetLength(FChannels[i].AttData, Length(FChannels[aOldValue - 1].AttData));
    for j:= 0 to High(FChannels[i].AttData) do
      FChannels[i].AttData[j]:= FChannels[aOldValue - 1].AttData[j];
  end;
end;

function TCalibrIn.GetChannels: Integer;
begin
  Result:= Length(FChannels);
end;

procedure TCalibrIn.SetAttNmb(aValue: Integer);
var i, j, aOldValue: Integer;
begin
  if High(FChannels) < 0 then
    Exit;
  if aValue < 0 then
    aValue:= 0;  
  aOldValue:= Length(FChannels[0].AttData);
  if aValue = aOldValue then
    Exit;
  for i:= 0 to High(FChannels) do with FChannels[i] do begin
    SetLength(AttData, aValue);
    if aValue <= AttIdx then
      AttIdx:= High(AttData);
    if aValue > aOldValue then for j:= aOldValue to aValue - 1 do begin
      AttData[j].OffSet:= 0;
      AttData[j].AttVal:= 1 shl j;
    end;
  end;
end;

function TCalibrIn.GetAttNmb: Integer;
begin
  if Length(FChannels) > 0 then
    Result:= Length(FChannels[0].AttData)
  else
    Result:= 0;
end;

procedure TCalibrIn.SetAtt(aChannel: Integer; aValue: Integer);
begin
  FChannels[aChannel].AttIdx:= aValue;
end;

function TCalibrIn.GetAtt(aChannel: Integer): Integer;
begin
  Result:= FChannels[aChannel].AttIdx;
end;

function TCalibrIn.Input(const aData: TDataMatrix): Integer;
var i, j: Integer;
begin
  for i:= 0 to Min(High(aData), High(FChannels)) do with FChannels[i] do
    with AttData[AttIdx] do for j:= 0 to High(aData[i]) do
      aData[i, j]:= (aData[i, j] - OffSet) * AttVal;
  Result:= inherited Input(aData);
end;

function TCalibrIn.InputOffSetCalibration(const aData: TDataMatrix): Integer;
var i, j: Integer;
begin
  if Assigned(FMonitorFunction) then
    FMonitorFunction(aData);
  for i:= 0 to Min(High(aData), High(FChannels)) do with FChannels[i] do
    with AttData[AttIdx] do begin
      OffSet:= 0;
      for j:= 0 to High(aData[i]) do begin
        OffSet:= OffSet + aData[i, j];
      end;
      OffSet:= OffSet / Length(aData[i]);
    end;
  Result:= 1;
end;

function TCalibrIn.InputAmplitudeCalibration(const aData: TDataMatrix): Integer;
var aPos, aNeg, aRefAmp, aOffSet: Extended;
    aPosCount, aNegCount: Cardinal;
    i, j: Integer;
begin
  Result:= 1;
  if Assigned(FMonitorFunction) then
    FMonitorFunction(aData);
  aRefAmp:= 1;
  for i:= 0 to Min(High(aData), High(FChannels)) do with FChannels[i] do begin
    aOffSet:= 0;
    for j:= 0 to High(aData[i]) do
      aOffSet:= aOffSet + aData[i, j];
    aOffSet:= aOffSet / Length(aData[i]);
    aPos:= 0;      aNeg:= 0;
    aPosCount:= 0; aNegCount:= 0;
    for j:= 0 to High(aData[i]) do begin
      aData[i,j]:= aData[i,j] - aOffSet;
      if aData[i,j] > 0 then begin
        aPos:= aPos + aData[i,j];
        Inc(aPosCount);
      end else begin
        aNeg:= aNeg + aData[i,j];
        Inc(aNegCount);
      end;
    end;
    if (aPosCount = 0) or (aNegCount = 0) then
      aPos:= 0
    else
      aPos:= (aPos / aPosCount) - (aNeg / aNegCount);
    if aPos <= 0 then begin
      if i <> FRefChannel then
        AttNumber:= AttIdx
      else
        if Assigned(FMessageProc) then
           FMessageProc('No signal on amplitude calibration.', False);
      Exit;
    end else
      if i = FRefChannel then
        aRefAmp:= AttData[AttIdx].AttVal * aPos
      else
        AttData[AttIdx].AttVal:= 1 / aPos;
  end;
  for i:= 0 to Min(High(aData), High(FChannels)) do if i <> FRefChannel then
    with FChannels[i] do
      AttData[AttIdx].AttVal:= AttData[AttIdx].AttVal * aRefAmp;
end;

procedure TCalibrIn.SaveIni(aFile: TIniFile);
var i, j: Integer;
  aSt: String;
begin
  aFile.WriteInteger(sCalibr, 'Channels', Channels);
  aFile.WriteInteger(sCalibr, 'AttNumber', AttNumber);
  for i:= 0 to High(FChannels) do with FChannels[i] do
      for j:= 0 to High(AttData) do with AttData[j] do begin
        aSt:= IntToStr(i)+IntToStr(j);
        aFile.WriteFloat(sCalibr, 'OffSet' + aSt, OffSet);
        aFile.WriteFloat(sCalibr, 'AttVal' + aSt, AttVal);
      end;
end;

procedure TCalibrIn.LoadIni(aFile: TIniFile);
var i, j: Integer;
  aSt: String;
begin
  Channels:= aFile.ReadInteger(sCalibr, 'Channels', Channels);
  AttNumber:= aFile.ReadInteger(sCalibr, 'AttNumber', AttNumber);
  for i:= 0 to High(FChannels) do with FChannels[i] do begin
    AttIdx:= 0;
    for j:= 0 to High(AttData) do with AttData[j] do begin
      aSt:= IntToStr(i)+IntToStr(j);
      OffSet:= aFile.ReadFloat(sCalibr, 'OffSet' + aSt, OffSet);
      AttVal:= aFile.ReadFloat(sCalibr, 'AttVal' + aSt, AttVal);
    end;
  end;
end;

end.
