unit uMixer;

interface

uses
  MMSystem, Classes, Windows,
  uMixerControls, uCommon;

type
  TMixer = class(TMixerControls)
  private
    FCaps: MIXERCAPS;
    FMixerLine: MixerLine;
    FControls: MIXERLINECONTROLS;
    FCtrlsArray: array of TMIXERCONTROL;
    FSources: array of record
                Name: String;
                VolIdx: Integer;
                Volumes: array of Double;
                MutIdx: Integer;
                MUXItm: DWORD;
              end;
    FMUXCtrlIdx: Integer;
    FMastVolIdx: Integer;
    FMastVolume: array of Double;
    FCurrentSource: Integer;
    FMixerChange: TNotifyEvent;
    function SetMixerID(aMixerID: UINT; aCmpType: DWORD): Boolean;
    function SourceInRange(aProcName: String = ''): Boolean;
    function SourceChannelInRange(aChannel: Integer; aProcName: String = ''): Boolean;

    procedure SetSource(aSource: Integer);
    procedure SetVolume(aChannel: Integer; aVolume: Double);
    function GetVolume(aChannel: Integer): Double;
    function GetSourceName(aSource: Integer): String;
  protected
    procedure Init; override;
  public
    constructor Create(AOwner: TComponent; aMessageProc: TMessageProc = nil); override;
    destructor Destroy; override;

    function SetInDevice(aDevice: Integer): Boolean;
    function SetOutDevice(aDevice: Integer): Boolean;

    function SourcesNumber: Integer;
    function MixerName: String;
    function MixChannels: Integer;

    property Source: Integer read FCurrentSource write SetSource;
    property Volume[aChannel: Integer]: Double read GetVolume write SetVolume;
    property SourceName[aSource: Integer]: String read GetSourceName;
    property OnChangeMixer: TNotifyEvent write FMixerChange;
  end;

implementation

uses Sysutils;

procedure TMixer.Init;
begin
  ZeroMemory(@FCaps, SizeOf(FCaps));
  FMUXCtrlIdx:= -1;
  FMastVolIdx:= -1;
  FCurrentSource:= -1;
  FSources:= nil;
  FMastVolume:= nil;
  inherited;
end;

constructor TMixer.Create(AOwner: TComponent; aMessageProc: TMessageProc = nil);
begin
  inherited;
  Init;
end;

destructor TMixer.Destroy;
begin
  Init;
  inherited;
end;

function TMixer.SetMixerID(aMixerID: UINT; aCmpType: DWORD): Boolean;
const csCaptured = ': Can not capture control.';
var
  i, j, k: Integer;
  aMastMutIdx: Integer;
  aMixBool: MIXERCONTROLDETAILS_BOOLEAN;
  aMixUns: MIXERCONTROLDETAILS_UNSIGNED;
  aDetails: TMIXERCONTROLDETAILS;
  aListText: array of MIXERCONTROLDETAILS_LISTTEXT;
  aLineInFound: Boolean;
  aProcName: String;
begin
  if aMixerID = FDeviceID then begin
    Result:= True;
    Exit;
  end;
  Result:= False;
  Init;
  FDeviceID:= aMixerID;
  aProcName:= 'Set device ' + IntToStr(FDeviceID);
  try
    FError:= mixerOpen(@FDeviceHD, FDeviceID, Handle, 0, CALLBACK_WINDOW);
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes(aProcName + ': Open: ' + SysErrMes);
      Exit;
    end;
    FIsOpen:= True;
    FError:= mixerGetDevCaps(FDeviceID, @FCaps, SizeOf(FCaps));
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes(aProcName + ': Caps: ' + SysErrMes);
      Exit;
    end;
    aProcName:= aProcName + ' "' + FCaps.szPname + '"';
    InfMes(aProcName + '...');
    aProcName:= aProcName + ': ';
    if FCaps.cDestinations = 0 then begin
      ErrMes(aProcName + 'Destinations not found.');
      Exit;
    end;
    ZeroMemory(@FMixerLine, SizeOf(FMixerLine));
    FMixerLine.cbStruct:= SizeOf(FMixerLine);
    for i:= 0 to FCaps.cDestinations - 1 do begin
      FMixerLine.dwDestination:= i;
      FMixerLine.dwSource:= 0;
      FError:= mixerGetLineInfo(FDeviceHD, @FMixerLine, MIXER_GETLINEINFOF_DESTINATION);
      if FError <> MMSYSERR_NOERROR then begin
        ErrMes(aProcName + 'DstLineInfo: Destination ' + IntToStr(i)
          + ': ' + SysErrMes);
        Exit;
      end;
      {$Q-}
      if FMixerLine.dwComponentType <> aCmpType then begin
        if LongWord(i + 1) = FCaps.cDestinations then begin
          ErrMes(aProcName + 'Required destination not found.');
          Exit;
        end else
          Continue;
      end;
      ZeroMemory(@FControls, SizeOf(FControls));
      if FMixerLine.cControls > 0 then begin
        SetLength(FCtrlsArray, FMixerLine.cControls);
        FControls.pamxctrl:= @FCtrlsArray[0];
        FControls.dwLineID:= FMixerLine.dwLineID;
        FControls.cControls:= FMixerLine.cControls;
        FControls.cbStruct:= SizeOf(FControls);
        FControls.cbmxctrl:= SizeOf(FCtrlsArray[0]);
        FError:= mixerGetLineControls(FDeviceHD, @FControls, MIXER_GETLINECONTROLSF_ALL);
        if FError <> MMSYSERR_NOERROR then begin
          ErrMes(aProcName + 'DstCtrlsAll: ' + LineInfo(@FMixerLine)
            + ': ' + SysErrMes);
          Exit;
        end;
        aMastMutIdx:= -1;
        for j:= 0 to FControls.cControls - 1 do
          case FCtrlsArray[j].dwControlType of
            MIXERCONTROL_CONTROLTYPE_MUX,
            MIXERCONTROL_CONTROLTYPE_MIXER: if FMUXCtrlIdx < 0 then begin
              aDetails.cbStruct:= SizeOf(aDetails);
              aDetails.dwControlID:= FCtrlsArray[j].dwControlID;
              aDetails.cChannels:= 1;
              aDetails.cMultipleItems:= FCtrlsArray[j].cMultipleItems;
              aDetails.cbDetails:= SizeOf(MIXERCONTROLDETAILS_LISTTEXT);
              SetLength(aListText, FCtrlsArray[j].cMultipleItems);
              aDetails.paDetails:= @aListText[0];
              FError:= mixerGetControlDetails(FDeviceHD, @aDetails, MIXER_GETCONTROLDETAILSF_LISTTEXT);
              if FError <> MMSYSERR_NOERROR then begin
                ErrMes(aProcName + 'DstCtrlDetListText: '
                  + CtrlInfo(@FCtrlsArray[j]) + ': ' + SysErrMes);
                Exit;
              end;
              FMUXCtrlIdx:= AddControl(@FCtrlsArray[j], FMixerLine.cChannels);
              if FMUXCtrlIdx < 0 then
                Exit;
              if not CaptureControl(FMUXCtrlIdx) then begin
                ErrMes(aProcName +  CtrlInfo(@FCtrlsArray[j]) + csCaptured);
                Exit;
              end;
            end;
            MIXERCONTROL_CONTROLTYPE_VOLUME: if FMastVolIdx < 0 then begin
              FMastVolIdx:= AddControl(@FCtrlsArray[j], FMixerLine.cChannels);
              if FMastVolIdx < 0 then
                Exit;
              SetLength(FMastVolume, FMixerLine.cChannels);
              if not CaptureControl(FMastVolIdx) then begin
                ErrMes(aProcName +  CtrlInfo(@FCtrlsArray[j]) + csCaptured);
                Exit;
              end;
              aMixUns.dwValue:= High(WORD);
              for k:= 0 to High(FMastVolume) do begin
                if not SetValue(FMastVolIdx, k, 0, @aMixUns) then begin
                  ErrMes(aProcName +  'Can not set volume for channel '
                    + IntToStr(k) + ' for control: '
                    + CtrlInfo(@FCtrlsArray[j]) + '.');
                  Exit;
                end;
                FMastVolume[k]:= 1;
              end;
            end;
            MIXERCONTROL_CONTROLTYPE_MUTE: if aMastMutIdx < 0 then begin
              aMastMutIdx:= AddControl(@FCtrlsArray[j], 1);
              if aMastMutIdx < 0 then
                Exit;
              if not CaptureControl(aMastMutIdx) then begin
                ErrMes(aProcName +  CtrlInfo(@FCtrlsArray[j]) + csCaptured);
                Exit;
              end;
              aMixBool.fValue:= 0;
              if not SetValue(aMastMutIdx, 0, 0, @aMixBool) then begin
                ErrMes(aProcName +  'Can not set value for control: '
                  + CtrlInfo(@FCtrlsArray[j]) + '.');
                Exit;
              end;
            end;
          end;
      end;
      Break;
    end;
    SetLength(FSources, FMixerLine.cConnections);
    aLineInFound:= False;
    for i:= 0 to High(FSources) do begin
      FMixerLine.dwSource:= i;
      FMixerLine.cbStruct:= SizeOf(FMixerLine);
      FError:= mixerGetLineInfo(FDeviceID, @FMixerLine, MIXER_GETLINEINFOF_SOURCE);
      if FError <> MMSYSERR_NOERROR then begin
        ErrMes(aProcName + 'SrcLineInfo ' + IntToStr(i) + ': ' + SysErrMes);
        Exit;
      end;
      FSources[i].Name:= FMixerLine.szShortName;
      FSources[i].VolIdx:= -1;
      FSources[i].Volumes:= nil;
      FSources[i].MutIdx:= -1;
      FSources[i].MUXItm:= High(DWORD);
      if FMUXCtrlIdx >= 0 then
        for j:= 0 to High(aListText) do
          if aListText[j].dwParam1 = FMixerLine.dwLineID then begin
            FSources[i].MUXItm:= j;
            Break;
          end;
      if FMixerLine.cControls > 0 then begin
        SetLength(FCtrlsArray, FMixerLine.cControls);
        FControls.pamxctrl:= @FCtrlsArray[0];
        FControls.dwLineID:= FMixerLine.dwLineID;
        FControls.cControls:= FMixerLine.cControls;
        FControls.cbStruct:= SizeOf(FControls);
        FControls.cbmxctrl:= SizeOf(FCtrlsArray[0]);
        FError:= mixerGetLineControls(FDeviceHD, @FControls, MIXER_GETLINECONTROLSF_ALL);
        if FError <> MMSYSERR_NOERROR then begin
          ErrMes(aProcName + 'SrcCtrlsAll: ' + LineInfo(@FMixerLine)
            + ': ' + SysErrMes);;
          Exit;
        end;
        for j:= 0 to FMixerLine.cControls - 1 do
          case FCtrlsArray[j].dwControlType of
            MIXERCONTROL_CONTROLTYPE_VOLUME: if FSources[i].VolIdx < 0 then begin
              FSources[i].VolIdx:= AddControl(@FCtrlsArray[j], FMixerLine.cChannels);
              if FSources[i].VolIdx < 0 then
                Exit;
              SetLength(FSources[i].Volumes, FMixerLine.cChannels);
            end;
            MIXERCONTROL_CONTROLTYPE_MUTE: if FSources[i].MutIdx < 0 then begin
              FSources[i].MutIdx:= AddControl(@FCtrlsArray[j], 1);
              if FSources[i].MutIdx < 0 then
                Exit;
              if not CaptureControl(FSources[i].MutIdx) then begin
                ErrMes(aProcName +  CtrlInfo(@FCtrlsArray[j]) + csCaptured);
                Exit;
              end;
              aMixBool.fValue:= 1;
              if not SetValue(FSources[i].MutIdx, 0, 0, @aMixBool) then begin
                ErrMes(aProcName + 'Can not set value for control: '
                  + CtrlInfo(@FCtrlsArray[j]) + '.');
                Exit;
              end;
            end;
          end;
       end;
       case aCmpType of
        MIXERLINE_COMPONENTTYPE_DST_SPEAKERS: if (FCurrentSource < 0) and
          (FMixerLine.dwComponentType = MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT) then
            SetSource(i);
        MIXERLINE_COMPONENTTYPE_DST_WAVEIN: if (FCurrentSource < 0) and (not aLineInFound) then
          case FMixerLine.dwComponentType of
            MIXERLINE_COMPONENTTYPE_SRC_LINE: begin
              SetSource(i);
              aLineInFound:= FCurrentSource >= 0;
            end;
            MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY:
              SetSource(i);
          end;
       end;
    end;
    if Assigned(FMixerChange) then
      FMixerChange(Self);
    InfMes(aProcName + 'Done.');
    Result:= True;  
  except
    ErrMes(aProcName + 'Exception.');
  end;
end;

function TMixer.SourcesNumber: Integer;
begin
  Result:= Length(FSources);
end;

function TMixer.MixerName: String;
begin
  Result:= FCaps.szPname;
end;

function TMixer.SourceInRange(aProcName: String = ''): Boolean;
begin
  Result:= (FCurrentSource >= 0) and (FCurrentSource < Length(FSources));
  if (not Result) and (aProcName <> '') then
    ErrMes(aProcName + ': Current source index is out of range.');
end;

function TMixer.MixChannels: Integer;
begin
  if SourceInRange then
    Result:= Length(FSources[FCurrentSource].Volumes)
  else if FMastVolIdx > 0 then
    Result:= Length(FMastVolume)
  else
    Result:= 0;
end;

function TMixer.SourceChannelInRange(aChannel: Integer; aProcName: String = ''): Boolean;
begin
{$B-}
  Result:= SourceInRange(aProcName);
  if not Result then
    Exit;
  Result:= (aChannel >= 0) and (aChannel < Length(FSources[FCurrentSource].Volumes));
  if (not Result) and (aProcName <> '') then
    ErrMes(aProcName + ': Channel index is out of range.');
end;

procedure TMixer.SetSource(aSource: Integer);
var
  aMixBool: MIXERCONTROLDETAILS_BOOLEAN;
  aProcName: String;
begin
  if aSource = FCurrentSource then
    Exit;
  aProcName:= 'Set source ' + IntToStr(aSource);
  if (High(FSources) < 0) or (aSource < 0) or
     (aSource > High(FSources)) then begin
    ErrMes(aProcName + ': Source index is out of range.');
    Exit;
  end;
  aProcName:= aProcName + ' "' + FSources[aSource].Name + '"';
  try
    if 0 <= FSources[aSource].VolIdx then
      if IsCaptured(FSources[aSource].VolIdx) then
        Exit;
    // Current source controls processing
    if (FCurrentSource >= 0) and (FCurrentSource < Length(FSources)) then begin
      with FSources[FCurrentSource] do begin
        if (VolIdx >= 0) then
          if not LeaveControl(VolIdx) then begin
            ErrMes(aProcName + ': Can not leave source Volume "'
              + Name + '".');
            Exit;
          end;
        aMixBool.fValue:= 1;
        if (MutIdx >= 0) then
          if not SetValue(MutIdx, 0, 0, @aMixBool) then begin
            ErrMes(aProcName + ': Can not set source Mute "'
              + Name + '".');
            Exit;
          end;  
      end;
    end;

    FCurrentSource:= aSource;

    // New source controls processing
    with FSources[FCurrentSource] do begin
      if FMUXCtrlIdx >= 0 then begin
        aMixBool.fValue:= 1;
        if not SetValue(FMUXCtrlIdx, 0, MUXItm, @aMixBool) then begin
          ErrMes(aProcName + ': Can not set lines Selector.');
          Exit;
        end;
      end else if MutIdx >= 0 then begin
        aMixBool.fValue:= 0;
        if not SetValue(MutIdx, 0, 0, @aMixBool) then begin
          ErrMes(aProcName + ': Can not set line Mute.');
          Exit;
        end;
      end;
      if (VolIdx >= 0) then
        if not CaptureControl(VolIdx) then begin
          ErrMes(aProcName + ': Can not capture Volume.');
          Exit;
        end;
    end;
    InfMes(aProcName + '.');
  except
    ErrMes(aProcName + ': Exeption.');
  end;
end;

procedure TMixer.SetVolume(aChannel: Integer; aVolume: Double);
var aMixUns: MIXERCONTROLDETAILS_UNSIGNED;
    aProcName: String;
begin
  if aVolume < 0 then
    aVolume:= 0;
  if aVolume > 1 then
    aVolume:= 1;
  aMixUns.dwValue:= Round(High(WORD) * aVolume);
  aProcName:= 'Set volume: Channel: ' + IntToStr(aChannel) +
              ',  Value: ' + FloatToStr(aVolume);
  try
    if SourceChannelInRange(aChannel) then begin
      if not SetValue(FSources[FCurrentSource].VolIdx, aChannel, 0, @aMixUns) then begin
        ErrMes(aProcName + ': Can not set value for source channel.');
        Exit;
      end;
      FSources[FCurrentSource].Volumes[aChannel]:= aVolume;
    end else if (FMastVolIdx >= 0) and (aChannel < Length(FMastVolume)) then begin
      if not SetValue(FMastVolIdx, aChannel, 0, @aMixUns) then begin
        ErrMes(aProcName + ': Can not set value for master channel.');
        Exit;
      end;
      FMastVolume[aChannel]:= aVolume;
    end else begin
      ErrMes(aProcName + ': Volume control not found.');
      Exit;
    end;
    InfMes(aProcName + '.');
  except
    ErrMes(aProcName + ': Exception.');
  end;
end;

function TMixer.GetVolume(aChannel: Integer): Double;
begin
  if SourceChannelInRange(aChannel) then
    Result:= FSources[FCurrentSource].Volumes[aChannel]
  else if (FMastVolIdx >= 0) and (aChannel < Length(FMastVolume)) then
    Result:= FMastVolume[aChannel]
  else begin
    ErrMes('GetVolume: Volume control not found.');
    Result:= 0;
  end;
end;

function TMixer.GetSourceName(aSource: Integer): String;
begin
  Result:= FSources[aSource].Name;
end;

function TMixer.SetInDevice(aDevice: Integer): Boolean;
var aMixID: UINT;
begin
  Result:= False;
  FObjectName:= 'MixerIn';
  FError:= mixerGetID(aDevice, aMixID, MIXER_OBJECTF_WAVEIN);
  if FError = MMSYSERR_NOERROR then
    Result:= SetMixerID(aMixID, MIXERLINE_COMPONENTTYPE_DST_WAVEIN)
  else begin
    ErrMes('SetWaveIn ' + IntToStr(aDevice) + ': GetMixID: ' + SysErrMes);
    Exit;
  end;
end;

function TMixer.SetOutDevice(aDevice: Integer): Boolean;
var aMixID: UINT;
begin
  Result:= False;
  FObjectName:= 'MixerOut';
  FError:= mixerGetID(aDevice, aMixID, MIXER_OBJECTF_WAVEOUT);
  if FError = MMSYSERR_NOERROR then
    Result:= SetMixerID(aMixID, MIXERLINE_COMPONENTTYPE_DST_SPEAKERS)
  else begin
    ErrMes('SetWaveOut ' + IntToStr(aDevice) + ': GetMixID: ' + SysErrMes);
    Exit;
  end;
end;

end.
