unit uSignalIn;

interface

uses
  Forms, Sysutils, Classes, IniFiles,
  uCommon, uWaveIn, uMixer, uCalibrIn;

type
  TSignalIn = class
  private
    //Classes
    FWaveIn: TWaveIn;
    FMixer: TMixer;
    FCalibrIn: TCalibrIn;

    // Wave
    FCount: LongWord;

    //Events
    FMessageProc: TMessageProc;
    FOnStart: TNotifyEvent;
    FOnStop: TNotifyEvent;

    //Internal
    procedure OnCloseCalibrOffSet(Sender: TObject);
    procedure OnCloseCalibrAmplitude(Sender: TObject);

    //Wave
    function GetDevicesNumber: Integer;
    function GetDeviceName(aDevice: Integer): String;
    procedure SetDevice(aDevice: Integer);
    function GetDevice: Integer;
    procedure SetBits(aValue: Word);
    function GetBits: Word;
    function GetSampling: LongWord;
    procedure SetSampling(aValue: LongWord);
    function GetRunning: Boolean;

    //Mixer
    function GetSourcesNumber: Integer;
    function GetSourceName(aSource: Integer): String;
    function GetSource: Integer;
    procedure SetSource(aSource: Integer);
    function GetAtt(aChannel: Integer): Integer;
    procedure SetAtt(aChannel, aValue: Integer);
    function GetAttNum: Integer;

    //Calibration
    procedure AddOutput(aOutput: TDataFunction);
    procedure SetMonitorFunction(aOutput: TDataFunction);

    //Events
    procedure SetMessageProc(aProc: TMessageProc);
    procedure SetOnChangeDevice(aEvent: TNotifyEvent);
    procedure SetOnSaturation(aFunc: TIntFunction);
    procedure SetOnLowSignal(aFunc: TIntFunction);
    procedure SetOnStop(aEvent: TNotifyEvent);
  public
    constructor Create(AOwner: TComponent; var aResult: Boolean; aMessageProc: TMessageProc = nil);
    destructor Destroy; override;

    //Actions
    procedure CalibrateOffSet;
    procedure CalibrateAmplitude(aFreq: Extended);
    procedure Start;
    procedure Stop;

    //Wave
    procedure InvalidateInput(Sender: TObject);
    property DevicesNumber: Integer read GetDevicesNumber;
    property DeviceName[aDevice: Integer]: String read GetDeviceName;
    property Device: Integer read GetDevice write SetDevice;
    property Count: LongWord read FCount write FCount;
    property Bits: Word read GetBits write SetBits;        
    property Sampling: LongWord read GetSampling write SetSampling;
    property Running: Boolean read GetRunning;

    //Mixer
    property Source: Integer read GetSource write SetSource;
    property SourceName[aSource: Integer]: String read GetSourceName;
    property SourcesNumber: Integer read GetSourcesNumber;
    procedure SaveIni(aFile: TIniFile);
    procedure LoadIni(aFile: TIniFile);

    //Calibration
    property Attenuator[aChannel: Integer]: Integer read GetAtt write SetAtt;
    property Output: TDataFunction write AddOutput;
    property AttNumber: Integer read GetAttNum;
    property MonitorFunction: TDataFunction write SetMonitorFunction;

    // Events
    property MessageProc: TMessageProc write SetMessageProc;
    property OnStop: TNotifyEvent write SetOnStop;
    property OnStart: TNotifyEvent write FOnStart;
    property OnChangeDevice: TNotifyEvent write SetOnChangeDevice;
    property OnSaturation: TIntFunction write SetOnSaturation;
    property OnLowSignal: TIntFunction write SetOnLowSignal;
  end;

implementation

uses Windows;

constructor TSignalIn.Create(AOwner: TComponent; var aResult: Boolean; aMessageProc: TMessageProc = nil); 
begin
  inherited Create;
  aResult:= False;
  FMessageProc:= aMessageProc;
  FWaveIn:= TWaveIn.Create(AOwner, aResult, aMessageProc);
  if not aResult then
    Exit;
  FCount:= FWaveIn.Count;
  if Assigned(aMessageProc) then
    aMessageProc('MixerIn: Creating ...', True);
  FMixer:= TMixer.Create(AOwner, aMessageProc);
  aResult:= FMixer.SetInDevice(FWaveIn.Device);
  if not aResult then
    Exit;
  if Assigned(aMessageProc) then
    aMessageProc('CalibrIn: Creating ...', True);
  FCalibrIn:= TCalibrIn.Create;
  FCalibrIn.MessageProc:= aMessageProc;
  FMixer.OnChangeControl:= FWaveIn.InvalidateInput;
  if Assigned(aMessageProc) then
    aMessageProc('SignalIn: Created.', True);
  aResult:= True;
end;

destructor TSignalIn.Destroy;
begin
  try
    FreeAndNil(FWaveIn);
    if Assigned(FMessageProc) then
      FMessageProc('WaveIn destroyed.', True);
  except
    if Assigned(FMessageProc) then
      FMessageProc('Can not destroy WaveIn.', False);
  end;
  try
    FreeAndNil(FMixer);
    if Assigned(FMessageProc) then
      FMessageProc('MixerIn destroyed.', True);
  except
    if Assigned(FMessageProc) then
      FMessageProc('Can not destroy MixerIn.', False);
  end;
  try
    FreeAndNil(FCalibrIn);
    if Assigned(FMessageProc) then
      FMessageProc('CalibrIn destroyed.', True);
  except
    if Assigned(FMessageProc) then
      FMessageProc('Can not destroy CalibrIn.', False);
  end;
  inherited;
end;

function TSignalIn.GetDevice: Integer;
begin
  Result:= FWaveIn.Device;
end;

procedure TSignalIn.SetDevice(aDevice: Integer);
begin
  FWaveIn.Device:= aDevice;
  FMixer.SetInDevice(FWaveIn.Device);
end;

procedure TSignalIn.Start;
begin
  FWaveIn.Count:= FCount;
  FWaveIn.DataFunction:= FCalibrIn.Input;
  FWaveIn.OnOpen:= FOnStart;
  FWaveIn.OnClose:= FOnStop;
  FWaveIn.Open;
end;

procedure TSignalIn.Stop;
begin
  FWaveIn.Close;
end;

procedure TSignalIn.AddOutput(aOutput: TDataFunction);
begin
  FCalibrIn.Output:= aOutput;
end;

procedure TSignalIn.SetMessageProc(aProc: TMessageProc);
begin
  FWaveIn.MessageProc:= aProc;
  FMixer.MessageProc:= aProc;
  FMessageProc:= aProc;
end;

function TSignalIn.GetAtt(aChannel: Integer): Integer;
begin
  Result:= FCalibrIn.Attenuator[aChannel];
end;

procedure TSignalIn.SetAtt(aChannel, aValue: Integer);
begin
  FCalibrIn.Attenuator[aChannel]:= aValue;
  FMixer.Volume[aChannel]:= 1 / (1 shl aValue);
end;

procedure TSignalIn.CalibrateOffSet;
var i: Integer;
begin
  if RelVer then
    FWaveIn.Count:= FWaveIn.Sampling div 5
  else
    FWaveIn.Count:= FWaveIn.Sampling div 100;
  FWaveIn.DataFunction:= FCalibrIn.InputOffSetCalibration;
  for i:= 0 to FCalibrIn.Channels - 1 do
    Attenuator[i]:= 0;
  FWaveIn.OnOpen:= FOnStart;
  FWaveIn.OnClose:= OnCloseCalibrOffSet;
  FWaveIn.Open;
end;

procedure TSignalIn.OnCloseCalibrOffSet(Sender: TObject);
var aAtt, i: Integer;
begin
  aAtt:= Attenuator[0];
  if aAtt < FCalibrIn.AttNumber - 1 then begin
    aAtt:= aAtt + 1;
    for i:= 0 to FCalibrIn.Channels - 1 do
      Attenuator[i]:= aAtt;
    FWaveIn.Open;
  end else if Assigned(FOnStop) then
    FOnStop(Self);
end;

procedure TSignalIn.CalibrateAmplitude(aFreq: Extended);
var i: Integer;
  cCalTime: Extended;  // Calibration time, approx. [sec]
begin
  if RelVer then
    cCalTime:= 0.5
  else
    cCalTime:= 0.1;
  i:= Round(cCalTime*aFreq);
  if i < 1 then i:= 1;
  {$Q-}
  FWaveIn.Count:= Round(i*FWaveIn.Sampling/aFreq);
  FWaveIn.DataFunction:= FCalibrIn.InputAmplitudeCalibration;
  for i:= 0 to FCalibrIn.Channels - 1 do
    Attenuator[i]:= 0;
  FWaveIn.OnOpen:= FOnStart;
  FWaveIn.OnClose:= OnCloseCalibrAmplitude;
  FCalibrIn.ReferenceChannel:= 0;
  FWaveIn.Open;
end;

procedure TSignalIn.OnCloseCalibrAmplitude(Sender: TObject);
var i: Integer;
begin
  if FCalibrIn.AttNumber = 1 then begin
    if Assigned(FOnStop) then
      FOnStop(Self);
    Exit;
  end;
  if FCalibrIn.ReferenceChannel = 0 then
    if Attenuator[1] < FCalibrIn.AttNumber - 1 then
      for i:= 1 to FCalibrIn.Channels - 1 do
        Attenuator[i]:= Attenuator[i] + 1
    else begin
      FCalibrIn.ReferenceChannel:= 1;
      Attenuator[1]:= 0;
      Attenuator[0]:= 1;
    end
  else
    if Attenuator[0] < FCalibrIn.AttNumber - 1 then
      Attenuator[0]:= Attenuator[0] + 1
    else begin
      if Assigned(FOnStop) then
        FOnStop(Self);
      Exit;
    end;
  FWaveIn.Open;
end;

function TSignalIn.GetSource: Integer;
begin
  Result:= FMixer.Source;
end;

procedure TSignalIn.SetSource(aSource: Integer);
begin
  FMixer.Source:= aSource;
end;

function TSignalIn.GetSourceName(aSource: Integer): String;
begin
  Result:= FMixer.SourceName[aSource];
end;

procedure TSignalIn.SetOnChangeDevice(aEvent: TNotifyEvent);
begin
  FMixer.OnChangeMixer:= aEvent;
end;

function TSignalIn.GetSourcesNumber: Integer;
begin
  Result:= FMixer.SourcesNumber;
end;

function TSignalIn.GetAttNum: Integer;
begin
  Result:= FCalibrIn.AttNumber;
end;

function TSignalIn.GetSampling: LongWord;
begin
  Result:= FWaveIn.Sampling;
end;

procedure TSignalIn.SetSampling(aValue: LongWord);
begin
  FWaveIn.Sampling:= aValue;
end;

function TSignalIn.GetRunning: Boolean;
begin
  Result:= FWaveIn.IsOpen;
end;

function TSignalIn.GetDevicesNumber: Integer;
begin
  Result:= FWaveIn.DevicesNumber;
end;

function TSignalIn.GetDeviceName(aDevice: Integer): String;
begin
  Result:= FWaveIn.DeviceName[aDevice];
end;

procedure TSignalIn.SetOnSaturation(aFunc: TIntFunction);
begin
  FWaveIn.OnSaturation:= aFunc;
end;

procedure TSignalIn.SetOnLowSignal(aFunc: TIntFunction);
begin
  FWaveIn.OnLowSignal:= aFunc;
end;

procedure TSignalIn.SetMonitorFunction(aOutput: TDataFunction);
begin
  FCalibrIn.MonitorFunction:= aOutput;
end;

procedure TSignalIn.InvalidateInput(Sender: TObject);
begin
  FWaveIn.InvalidateInput(Sender);
end;

procedure TSignalIn.SetOnStop(aEvent: TNotifyEvent);
begin
  FOnStop:= aEvent;
  FWaveIn.OnClose:= aEvent;
end;

procedure TSignalIn.SaveIni(aFile: TIniFile);
begin
  FCalibrIn.SaveIni(aFile);
end;

procedure TSignalIn.LoadIni(aFile: TIniFile);
begin
  FCalibrIn.LoadIni(aFile);
end;

procedure TSignalIn.SetBits(aValue: Word);
begin
  FWaveIn.Bits:= aValue;
end;

function TSignalIn.GetBits: Word;
begin
  Result:= FWaveIn.Bits;
end;

end.
