unit uSignalOut;

interface

uses
  Forms, Sysutils, Classes,
  uCommon, uWaveOut, uMixer;

type
  TSignalOut = class
  private
    //Classes
    FWaveOut: TWaveOut;
    FMixer: TMixer;

    //Events
    FMessageProc: TMessageProc;
    FOnStart: TNotifyEvent;
    FOnStop: TNotifyEvent;

    //Internal

    //Wave
    function GetDevicesNumber: Integer;
    function GetDeviceName(aDevice: Integer): String;
    procedure SetDevice(aDevice: Integer);
    function GetDevice: Integer;
    function GetSampling: LongWord;
    procedure SetSampling(aValue: LongWord);
    function GetRunning: Boolean;
    function GetCount: LongWord;
    procedure SetBits(aValue: Word);
    function GetBits: Word;

    //Mixer
    function GetSourcesNumber: Integer;
    function GetSourceName(aSource: Integer): String;
    function GetSource: Integer;
    procedure SetSource(aSource: Integer);
    function GetAtt(aChannel: Integer): Integer;
    procedure SetAtt(aChannel, aValue: Integer);
    function GetChannels: Integer;
    procedure SetInvalidateOutput(aEvent: TNotifyEvent);

    //Events
    procedure SetMessageProc(aProc: TMessageProc);
    procedure SetOnChangeDevice(aEvent: TNotifyEvent);

  public
    constructor Create(AOwner: TComponent; var aResult: Boolean; aMessageProc: TMessageProc = nil);
    destructor Destroy; override;

    //Actions
    procedure Trapezium(aSlopeTime, aFlatTime: Extended);
    procedure MultiSin(aFreqNum, aStep: Integer; aCount: LongWord=1024);
    procedure Stop;

    //Wave
    property DevicesNumber: Integer read GetDevicesNumber;
    property DeviceName[aDevice: Integer]: String read GetDeviceName;
    property Device: Integer read GetDevice write SetDevice;
    property Bits: Word read GetBits write SetBits;
    property Sampling: LongWord read GetSampling write SetSampling;
    property Count: LongWord read GetCount;
    property Running: Boolean read GetRunning;

    //Mixer
    property Source: Integer read GetSource write SetSource;
    property SourceName[aSource: Integer]: String read GetSourceName;
    property SourcesNumber: Integer read GetSourcesNumber;
    property Attenuator[aChannel: Integer]: Integer read GetAtt write SetAtt;
    property Channels: Integer read GetChannels;

    // Events
    property MessageProc: TMessageProc write SetMessageProc;
    property OnStop: TNotifyEvent write FOnStop;
    property OnStart: TNotifyEvent write FOnStart;
    property OnChangeDevice: TNotifyEvent write SetOnChangeDevice;
    property OnInvalidateOutput: TNotifyEvent write SetInvalidateOutput;
  end;


implementation

uses
  Math, Windows;

constructor TSignalOut.Create(AOwner: TComponent; var aResult: Boolean; aMessageProc: TMessageProc = nil);
begin
  inherited Create;
  aResult:= False;
  FMessageProc:= aMessageProc;
  FWaveOut:= TWaveOut.Create(AOwner, aResult, aMessageProc);
  if not aResult then
    Exit;
  if Assigned(aMessageProc) then
    aMessageProc('MixerOut: Creating ...', True);
  FWaveOut.MessageProc:= aMessageProc;
  FMixer:= TMixer.Create(AOwner, aMessageProc);
  aResult:= FMixer.SetOutDevice(FWaveOut.Device);
  if not aResult then
    Exit;
  if Assigned(aMessageProc) then
    aMessageProc('SignalOut: Created.', True);    
  aResult:= True;
end;

destructor TSignalOut.Destroy;
begin
  try
    FreeAndNil(FWaveOut);
    if Assigned(FMessageProc) then
      FMessageProc('WaveOut destroyed.', True);
  except
    if Assigned(FMessageProc) then
      FMessageProc('Can not destroy WaveOut.', False);
  end;
  try
    FreeAndNil(FMixer);
    if Assigned(FMessageProc) then
      FMessageProc('MixerOut destroyed.', True)
  except
    FMessageProc('Can not destroy MixerOut.', False);
  end;
  inherited;
end;

function TSignalOut.GetSource: Integer;
begin
  Result:= FMixer.Source;
end;

procedure TSignalOut.SetSource(aSource: Integer);
begin
  FMixer.Source:= aSource;
end;

function TSignalOut.GetSourceName(aSource: Integer): String;
begin
  Result:= FMixer.SourceName[aSource];
end;

procedure TSignalOut.SetOnChangeDevice(aEvent: TNotifyEvent);
begin
  FMixer.OnChangeMixer:= aEvent;
end;

function TSignalOut.GetSourcesNumber: Integer;
begin
  Result:= FMixer.SourcesNumber;
end;

function TSignalOut.GetSampling: LongWord;
begin
  Result:= FWaveOut.Sampling;
end;

procedure TSignalOut.SetSampling(aValue: LongWord);
begin
  FWaveOut.Sampling:= aValue;
end;

function TSignalOut.GetRunning: Boolean;
begin
  Result:= FWaveOut.IsOpen;
end;

function TSignalOut.GetDevicesNumber: Integer;
begin
  Result:= FWaveOut.DevicesNumber;
end;

function TSignalOut.GetDeviceName(aDevice: Integer): String;
begin
  Result:= FWaveOut.DeviceName[aDevice];
end;

procedure TSignalOut.SetDevice(aDevice: Integer);
begin
  FWaveOut.Device:= aDevice;
  FMixer.SetOutDevice(FWaveOut.Device);
end;

function TSignalOut.GetDevice: Integer;
begin
 Result:= FWaveOut.Device;
end;

function TSignalOut.GetAtt(aChannel: Integer): Integer;
var aVolume: Extended;
begin
  aVolume:= FMixer.Volume[aChannel];
  if aVolume = 0 then
    Result:= 255
  else
    Result:= Round(Log2(1 / FMixer.Volume[aChannel]));
end;

procedure TSignalOut.SetAtt(aChannel, aValue: Integer);
begin
  FMixer.Volume[aChannel]:= 1 / (1 shl aValue);
end;

procedure TSignalOut.SetMessageProc(aProc: TMessageProc);
begin
//  FWaveOut.MessageProc:= aProc;
  FMixer.MessageProc:= aProc;
  FMessageProc:= aProc;
end;

procedure TSignalOut.Trapezium(aSlopeTime, aFlatTime: Extended);
begin
  FWaveOut.Trapezium(aSlopeTime, aFlatTime);
end;

procedure TSignalOut.MultiSin(aFreqNum, aStep: Integer; aCount: LongWord=1024);
begin
  FWaveOut.MultiSin(aFreqNum, aStep, aCount);
end;

procedure TSignalOut.Stop;
begin
  FWaveOut.Close;
end;

function TSignalOut.GetCount: LongWord;
begin
  Result:= FWaveOut.Count;
end;

function TSignalOut.GetChannels: Integer;
begin
  Result:= FMixer.MixChannels;
end;

procedure TSignalOut.SetInvalidateOutput(aEvent: TNotifyEvent);
begin
  FMixer.OnChangeControl:= aEvent;
end;

procedure TSignalOut.SetBits(aValue: Word);
begin
  FWaveOut.Bits:= aValue;
end;

function TSignalOut.GetBits: Word;
begin
  Result:= FWaveOut.Bits;
end;

end.
