unit uWaveIn;

interface

uses
  uCommon,
  MMSystem, Classes, Messages;

type
  TWaveIn = class(TMMWinControl)
  private
    FCaps: TWaveInCaps;
    FFormatEx: TWaveFormatEx;
    FMMTime: MMTIME;
    FFalseCount: LongWord;
    FFalseWHdr: TWaveHdr;
    FCount: LongWord;
    FWHdr: array of TWaveHdr;
    FBuffers: LongWord;
    FLocked: LongWord;
    FMaxSample: LongWord;
    FDataFunction: TDataFunction;
    FStartQuery: Boolean;
    FCloseQuery: Boolean;
    FQueryOpen: Boolean;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FOnInvalidate: TNotifyEvent;
    FOnSaturation: TIntFunction;
    FOnLowSignal: TIntFunction;
    FFunc: function(var a): Extended;
    FOutData: TDataMatrix;
    FOpenFlags: LongWord;
    function SetMaxCapsSettings: Boolean;
    procedure SetChannels(aValue: Word);
    function GetChannels: Word;
    procedure SetSampling(aValue: LongWord);
    function GetSampling: LongWord;
    procedure SetBits(aValue: Word);
    function GetBits: Word;
    procedure SetCount(aValue: LongWord);
    function GetDeviceName(aDevice: LongWord): String;
    procedure SetDevice(aDevice: LongWord);
    function AllocBuffers: Boolean;
    function AddBuffers: Boolean;
    function Start: Boolean;
    procedure OnWaveOpen(var Msg : TMessage); message MM_WIM_OPEN;
    procedure OnWaveIn(var Msg : TMessage); message MM_WIM_DATA;
    procedure OnWaveClose(var Msg : TMessage); message MM_WIM_CLOSE;
    function SysErrMes: String;
    function WaveFormat: String;
  public
    constructor Create(AOwner: TComponent; var aResult: Boolean;
                 aMessageProc: TMessageProc = nil); reintroduce;
    destructor Destroy; override;
    function DevicesNumber: Integer;
    function CheckDevice(aDevice: LongWord): Boolean;
    function CheckFormat: Boolean;
    procedure Open;
    procedure Close;
    procedure InvalidateInput(Sender: TObject);
    property Device: LongWord read FDeviceID write SetDevice;
    property DeviceName[aDevice: LongWord]: String read GetDeviceName;
    property Channels: Word read GetChannels write SetChannels;
    property Sampling: LongWord read GetSampling write SetSampling;
    property Bits: Word read GetBits write SetBits;
    property Count: LongWord read FCount write SetCount;
    property DataFunction: TDataFunction write FDataFunction;
    property OnOpen: TNotifyEvent write FOnOpen;
    property OnClose: TNotifyEvent write FOnClose;
    property OnInvalidate: TNotifyEvent write FOnInvalidate;
    property IsOpen: Boolean read FIsOpen;
    property OnSaturation: TIntFunction write FOnSaturation;
    property OnLowSignal: TIntFunction write FOnLowSignal;
  end;

implementation

uses
  Forms, Windows, Sysutils;

const
  cClassName = 'WaveIn';

constructor TWaveIn.Create(AOwner: TComponent; var aResult: Boolean;
                           aMessageProc: TMessageProc = nil);
var i: LongWord;
begin
  inherited Create(AOwner, aMessageProc);
  FObjectName:= 'WaveIn';
  aResult:= False;
  InfMes('Creating ...');
  FLocked:= 0;
  FDataFunction:= nil;
  FWHdr:= nil;
  FOnOpen:= nil;
  FOnClose:= nil;
  FOnInvalidate:= nil;
  FIsOpen:= False;
  FCloseQuery:= False;
  ZeroMemory(@FMMTime, SizeOf(FMMTime));
  FMMTime.wType:= TIME_SAMPLES;
  FBuffers:= 2;
  FCount:= 1024;
  FDeviceID:= High(FDeviceID);
  for i:= 0 to waveInGetNumDevs - 1 do if CheckDevice(i) then begin
    FDeviceID:= i;
    Break;
  end;
  if FDeviceID = High(FDeviceID) then
    ErrMes('Devices not found.')
  else
    aResult:= SetMaxCapsSettings;
end;

destructor TWaveIn.Destroy;
begin
  Close;
  while FIsOpen do
    Application.ProcessMessages;
  FWHdr:= nil;
  inherited;
end;

function TWaveIn.CheckDevice(aDevice: LongWord): Boolean;
begin
  ZeroMemory(@FCaps, SizeOf(FCaps));
  Result:= (MMSYSERR_NOERROR = waveInGetDevCaps(aDevice, @FCaps, SizeOf(FCaps)))
       and (FCaps.dwFormats <> WAVE_INVALIDFORMAT);
end;

function TWaveIn.CheckFormat: Boolean;
begin
  Result:= False;
  UpdateFormatEx(FFormatEx);
  FQueryOpen:= True;
  FError:= waveInOpen(@FDeviceHD, FDeviceID, @FFormatEx, Handle, 0, FOpenFlags);
  if FError = MMSYSERR_NOERROR then begin
    repeat Application.ProcessMessages
    until not FQueryOpen;
    Result:= True;
  end;
end;

function TWaveIn.WaveFormat: String;
begin
  Result:= 'Channels=' + IntToStr(Channels) +
    ', Sampling=' + IntToStr(Sampling) + ', Bits=' + IntToStr(Bits);
end;

function TWaveIn.SetMaxCapsSettings: Boolean;
var aOpenFlags, aMaxSampling: LongWord;
begin
  Result:= False;
  if FIsOpen or not CheckDevice(FDeviceID) then
    Exit;
  ZeroMemory(@FCaps, SizeOf(FCaps));
  waveInGetDevCaps(FDeviceID, @FCaps, SizeOf(FCaps));
  SetMaxFormatEx(FCaps.dwFormats, FFormatEx);
  FOpenFlags:= CALLBACK_WINDOW;
  if not CheckFormat then begin
    ErrMes('"' + DeviceName[FDeviceID] +
      '": Invalid wave format: ' + WaveFormat + '.');
    Exit;
  end;
  Result:= True;
  aOpenFlags:= FOpenFlags;
  with FFormatEx do begin
    if nSamplesPerSec = 44100 then begin
      nSamplesPerSec:= 48000;
      if not CheckFormat then
        nSamplesPerSec:= 44100
      else begin
        FOpenFlags:= aOpenFlags or WAVE_FORMAT_DIRECT;
        if not CheckFormat then
          FOpenFlags:= aOpenFlags;
        repeat
          aMaxSampling:= nSamplesPerSec;
          nSamplesPerSec:= aMaxSampling * 2;
          if nSamplesPerSec > cMaxSampling then begin
            aMaxSampling:= 48000;
            Break;
          end;
        until not CheckFormat;
        nSamplesPerSec:= aMaxSampling;
      end;
    end;
    InfMes('"' + DeviceName[FDeviceID] + '": Set format: ' + WaveFormat + '.');
  end;
end;

procedure TWaveIn.SetDevice(aDevice: LongWord);
begin
  if FIsOpen or not CheckDevice(aDevice) then
    Exit;
  FDeviceID:= aDevice;
  SetMaxCapsSettings;
end;

function TWaveIn.DevicesNumber: Integer;
begin
  Result:= waveInGetNumDevs;
end;

function TWaveIn.GetDeviceName(aDevice: LongWord): String;
begin
  waveInGetDevCaps(aDevice, @FCaps, SizeOf(FCaps));
  Result:= FCaps.szPname;
end;

procedure TWaveIn.SetChannels(aValue: Word);
var oldValue: Word;
begin
  if (aValue < 1) or FIsOpen or not CheckDevice(Device) then
    Exit;
  oldValue:= FFormatEx.nChannels;
  FFormatEx.nChannels:= aValue;
  if not CheckFormat then
    FFormatEx.nChannels:= oldValue;
end;

function TWaveIn.GetChannels: Word;
begin
  Result:= FFormatEx.nChannels;
end;

procedure TWaveIn.SetSampling(aValue: LongWord);
//var oldValue: LongWord;
begin
  if FIsOpen or not CheckDevice(Device) then
    Exit;
//  oldValue:= FFormatEx.nSamplesPerSec;
  FFormatEx.nSamplesPerSec:= aValue;
//  if not CheckFormat then
//    FFormatEx.nSamplesPerSec:= oldValue;
end;

function TWaveIn.GetSampling: LongWord;
begin
  Result:= FFormatEx.nSamplesPerSec;
end;

procedure TWaveIn.SetBits(aValue: Word);
//var oldValue: Word;
begin
  if FIsOpen or not CheckDevice(Device) then
    Exit;
  //oldValue:= FFormatEx.wBitsPerSample;
  FFormatEx.wBitsPerSample:= aValue;
  //if not CheckFormat then
  //  FFormatEx.wBitsPerSample:= oldValue;
end;

function TWaveIn.GetBits: Word;
begin
  Result:= FFormatEx.wBitsPerSample;
end;

procedure TWaveIn.SetCount(aValue: LongWord);
begin
  if FIsOpen then
    Exit;
  FCount:= aValue;
end;

function TWaveIn.AllocBuffers: Boolean;
var i: Integer;
begin
  Result:= False;
  FFalseCount:= 100 + Sampling div 10; // 100 + 0.1 sec
  with FFalseWHdr do begin
    ZeroMemory(@FFalseWHdr, SizeOf(FFalseWHdr));
    dwBufferLength:= FFalseCount * FFormatEx.nBlockAlign;
    lpData:= PChar(GlobalAlloc(AllocFlags, dwBufferLength));
    if lpData = nil then begin
      ErrMes('Can not allocate false buffer.');
      Exit;
    end;
  end;
  FBuffers:= 2 + Round(1.0{sec} * Sampling / FCount);
  SetLength(FWHdr, FBuffers);
  for i:= Low(FWHdr) to High(FWHdr) do with FWHdr[i] do begin
    ZeroMemory(@FWHdr[i], SizeOf(FWHdr[i]));
    dwBufferLength:= FCount * FFormatEx.nBlockAlign;
    lpData:= PChar(GlobalAlloc(AllocFlags, dwBufferLength));
    if lpData = nil then begin
      ErrMes('Can not allocate buffer #' + IntToStr(i) + '.');
      Exit;
    end;
  end;
  Result:= True;
end;

function TWaveIn.AddBuffers: Boolean;
var i: Integer;
begin
  {$Q-}
  Result:= False;
  FError:= waveInAddBuffer(FDeviceHD, @FFalseWHdr, SizeOf(FFalseWHdr));
  if FError <> MMSYSERR_NOERROR then begin
    ErrMes('Add false buffer: ' + SysErrMes);
    Exit;
  end;
  Inc(FMaxSample, FFalseCount);
  for i:= Low(FWHdr) to High(FWHdr) do begin
    FError:= waveInAddBuffer(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Add buffer #' + IntToStr(i) + ': ' + SysErrMes);
      Exit;
    end;
    Inc(FMaxSample, FCount);
    Inc(FLocked);
  end;
  Result:= True;
end;

procedure TWaveIn.Open;
begin
  if (FDeviceID = High(FDeviceID)) or FIsOpen then
    Exit;
  UpdateFormatEx(FFormatEx);
  FQueryOpen:= False;
  FError:= waveInOpen(@FDeviceHD, FDeviceID, @FFormatEx, Handle, 0, FOpenFlags);
  if FError <> MMSYSERR_NOERROR then begin
    ErrMes('Open: ' + SysErrMes);
    Exit;
  end;
end;

procedure TWaveIn.OnWaveOpen(var Msg : TMessage);
var i: Integer;
begin
  {$Q-}
  if FQueryOpen then begin
    waveInClose(FDeviceHD);
    Exit;
  end;
  FIsOpen:= True;
  if not AllocBuffers then
    Exit;
  with FFalseWHdr do begin
    FError:= waveInPrepareHeader(FDeviceHD, @FFalseWHdr, SizeOf(FFalseWHdr));
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Prepare false header: ' + SysErrMes);
      Exit;
    end;
  end;
  FMaxSample:= - 2 - Sampling shr 8;
  for i:= Low(FWHdr) to High(FWHdr) do with FWHdr[i] do begin
    FError:= waveInPrepareHeader(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Prepare header #' + IntToStr(i) + ': ' + SysErrMes);
      Exit;
    end;
  end;
  SetLength(FOutData, Channels, FCount);
  case Bits of
    8: FFunc:= Func8;
   16: FFunc:= Func16;
  else
    FFunc:= Func16;
  end;
  if Start then begin
    if Assigned(FOnOpen) then
      FOnOpen(Self);
    InfMes('Start: ' + WaveFormat + '.');
  end else
    Close;
end;

function TWaveIn.Start: Boolean;
begin
  Result:= False;
  if not AddBuffers then
    Exit;
  FError:= waveInStart(FDeviceHD);
  if FError = MMSYSERR_NOERROR then
    Result:= True
  else
    ErrMes('Start: ' + SysErrMes);
end;

procedure TWaveIn.OnWaveIn(var Msg : TMessage);
const
  cSemiLong = High(LongWord) div 2;
var
  WaveHeader: pWAVEHDR;
  hInputDev: HWAVEIN;
  aMaxSample,
  aCurSample: Int64;
  i, j, aResult: Integer;
  aBytes, aChannels: Integer;
  aData: PChar;
  aValue: Extended;
begin
  {$Q-}
  hInputDev:= HWAVEIN(Msg.WParam);
  WaveHeader:= pWAVEHDR(Msg.lParam);
  if WaveHeader = @FFalseWHdr then
    Exit;
  Dec(FLocked);
  if FStartQuery or FCloseQuery then
    if FLocked = 0 then begin
      if FStartQuery then begin
        FStartQuery:= False;
        Start;
      end;
      if FCloseQuery then begin
        FCloseQuery:= False;
        Close;
      end
    end else
  else
    if Assigned(FDataFunction) then begin
      aBytes:= Bits shr 3;
      aChannels:= Channels;
      aData:= WaveHeader.lpData;
      for i:= 0 to High(FOutData) do
        for j:= 0 to High(FOutData[i]) do begin
          FOutData[i, j]:= FFunc(aData[aBytes*(i+aChannels*j)]);
        end;
      aResult:= 0;
      for i:= 0 to High(FOutData) do begin
        aValue:= 0;
        for j:= 0 to High(FOutData[i]) do
          if Abs(FOutData[i, j]) >= aValue then
            aValue:= Abs(FOutData[i, j]);
        if (aValue > 0.99) and Assigned(FOnSaturation) then
          aResult:= FOnSaturation(i);
        if (aValue < 0.45) and Assigned(FOnLowSignal) then
          aResult:= FOnLowSignal(i);
      end;
      if aResult = 0 then
        aResult:= FDataFunction(FOutData)
      else
        aResult:= 0;
      case aResult of
        0: if FLocked = 0 then begin // Continue
             AddBuffers;
             if Assigned(FOnInvalidate) then
               FOnInvalidate(Self);
           end else begin
             if FStartQuery then
               Exit;
             waveInGetPosition(hInputDev, @FMMTime, SizeOf(FMMTime));
             aCurSample:= FMMTime.sample;
             aMaxSample:= FMaxSample;
             if Abs(aCurSample - aMaxSample) > cSemiLong then begin
               aCurSample:= FMMTime.sample + cSemiLong;
               aMaxSample:= FMaxSample + cSemiLong;
             end;
             if aCurSample > aMaxSample then begin
               FStartQuery:= True;
               if Assigned(FOnInvalidate) then
                 FOnInvalidate(Self);
             end else begin
               waveInAddBuffer(hInputDev, WaveHeader, SizeOf(WaveHeader^));
               Inc(FLocked);
               Inc(FMaxSample, FCount);
             end;
           end;
        else if FLocked = 0 then begin  //Close
          Close;
        end else begin
          FCloseQuery:= True;
          waveInReset(hInputDev);
        end;
      end // Case
    end else
      Close;
end;

procedure TWaveIn.Close;
var i: Integer;
begin
  if FCloseQuery then
    Exit;
  if FLocked > 0 then begin
    FCloseQuery:= True;
    FError:= waveInReset(FDeviceHD);
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Close: Reset: ' + SysErrMes);
      Exit;
    end;
  end else if FIsOpen then begin
    for i:= Low(FWHdr) to High(FWHdr) do with FWHdr[i] do begin
      FError:= waveInUnPrepareHeader(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
      if FError <> MMSYSERR_NOERROR then begin
        ErrMes('Unprepare header #' + IntToStr(i) + ': ' + SysErrMes);
        Exit;
      end;
      lpData:= Pointer(GlobalFree(THandle(lpData)));
      if lpData <> nil then begin
        ErrMes('Can not free handle #' + IntToStr(i) + '.');
        Exit;
      end;
    end;
    with FFalseWHdr do begin
      FError:= waveInUnPrepareHeader(FDeviceHD, @FFalseWHdr, SizeOf(FFalseWHdr));
      if FError <> MMSYSERR_NOERROR then begin
        ErrMes('Unprepare false header: ' + SysErrMes);
        Exit;
      end;
      lpData:= Pointer(GlobalFree(THandle(lpData)));
      if lpData <> nil then begin
        ErrMes('Can not free false handle.');
        Exit;
      end;
    end;
    FCloseQuery:= False;
    FError:= waveInClose(FDeviceHD);
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Close: ' + SysErrMes);
      Exit;
    end;
  end;
end;

procedure TWaveIn.OnWaveClose(var Msg : TMessage);
begin
  if FQueryOpen then begin
    FQueryOpen:= False;
    Exit;
  end;
  FIsOpen:= False;
  InfMes('Stop.');
  if Assigned(FOnClose) then
    FOnClose(Self);
end;

procedure TWaveIn.InvalidateInput(Sender: TObject);
begin
  if FLocked = 0 then
    Exit;
  FError:= waveInReset(FDeviceHD);
  if FError <> MMSYSERR_NOERROR then begin
    ErrMes('Invalidate: Reset: ' + SysErrMes);
    Exit;
  end;
  FStartQuery:= True;
  if Assigned(FOnInvalidate) then
    FOnInvalidate(Self);
end;

function TWaveIn.SysErrMes: String;
var aErrorText: array[1..MAXERRORLENGTH] of Char;
begin
  if FError = MMSYSERR_NOERROR then
    Result:= ''
  else begin
    waveInGetErrorText(FError, @aErrorText, SizeOf(aErrorText));
    Result:= aErrorText;
  end;
end;

end.
