unit uWaveOut;

interface

uses
  uCommon,
  MMSystem, Classes, Messages;

type
  TWaveOut = class(TMMWinControl)
  private
    FCaps: WaveOutCaps;
    FFormatEx: TWaveFormatEx;
    FCount: LongWord;
    FWHdr: array of TWaveHdr;
    FDataMatrix: TDataMatrix;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FLocked: LongWord;
    FCloseQuery: Boolean;
    FQueryOpen: Boolean;
    FOpenFlags: LongWord;    
    function SetMaxCapsSettings: Boolean;
    procedure SetChannels(aValue: Word);
    function GetChannels: Word;
    procedure SetSampling(aValue: LongWord);
    function GetSampling: LongWord;
    procedure SetBits(aValue: Word);
    function GetBits: Word;
    procedure SetCount(aValue: LongWord);
    function GetDeviceName(aDevice: LongWord): String;
    procedure SetDevice(aDevice: LongWord);
    procedure OnWaveOpen(var Msg : TMessage); message MM_WOM_OPEN;
    procedure OnWaveDone(var Msg : TMessage); message MM_WOM_DONE;
    procedure OnWaveClose(var Msg : TMessage); message MM_WOM_CLOSE;
    function SysErrMes: String;
    function WaveFormat: String;
    function GetDataMatrix(aChannel, aIndex: Integer): Extended;
    function AllocBuffers: Boolean;
  public
    constructor Create(AOwner: TComponent; var aResult: Boolean;
                 aMessageProc: TMessageProc = nil); reintroduce;
    destructor Destroy; override;
    function CheckDevice(aDevice: LongWord): Boolean;
    function CheckFormat: Boolean;
    function DevicesNumber: Integer;
    procedure Trapezium(aSlopeTime, aFlatTime: Extended);
    procedure MultiSin(aFreqNum, aStep: Integer; aCount: LongWord);
    procedure Open;
    procedure Close;
    property Device: LongWord read FDeviceID write SetDevice;
    property DeviceName[aDevice: LongWord]: String read GetDeviceName;
    property Channels: Word read GetChannels write SetChannels;
    property Sampling: LongWord read GetSampling write SetSampling;
    property Bits: Word read GetBits write SetBits;
    property Count: LongWord read FCount write SetCount;
    property Data[aIndex, aChannel: Integer]: Extended read GetDataMatrix;
    property OnOpen: TNotifyEvent write FOnOpen;
    property OnClose: TNotifyEvent write FOnClose;
    property IsOpen: Boolean read FIsOpen;
  end;

implementation

uses
  Forms, Windows, Math, Sysutils;

const
  cAmp = 0.45;

constructor TWaveOut.Create(AOwner: TComponent; var aResult: Boolean;
                      aMessageProc: TMessageProc = nil);
var i: LongWord;
begin
  inherited Create(AOwner, aMessageProc);
  FObjectName:= 'WaveOut';
  aResult:= False;
  InfMes('Creating ...');
  FLocked:= 0;
  FWHdr:= nil;
  FIsOpen:= False;
  FCount:= 1024;
  FDeviceID:= High(FDeviceID);
  for i:= 0 to waveOutGetNumDevs - 1 do if CheckDevice(i) then begin
    FDeviceID:= i;
    Break;
  end;
  if FDeviceID = High(FDeviceID) then
    ErrMes('Devices not found.')
  else
    aResult:= SetMaxCapsSettings;
end;

destructor TWaveOut.Destroy;
begin
  Close;
  while FIsOpen
  do Application.ProcessMessages;
  FWHdr:= nil; 
  inherited;
end;

function TWaveOut.CheckDevice(aDevice: LongWord): Boolean;
begin
  ZeroMemory(@FCaps, SizeOf(FCaps));
  Result:= (MMSYSERR_NOERROR = waveOutGetDevCaps(aDevice, @FCaps, SizeOf(FCaps)))
       and (FCaps.dwFormats <> WAVE_INVALIDFORMAT);
end;

procedure TWaveOut.SetDevice(aDevice: LongWord);
begin
  if FIsOpen or not CheckDevice(aDevice) then
    Exit;
  FDeviceID:= aDevice;
  SetMaxCapsSettings;
end;

function TWaveOut.CheckFormat: Boolean;
begin
  Result:= False;
  UpdateFormatEx(FFormatEx);
  FQueryOpen:= True;
  FError:= waveOutOpen(@FDeviceHD, FDeviceID, @FFormatEx, Handle, 0, FOpenFlags);
  if FError = MMSYSERR_NOERROR then begin
    repeat Application.ProcessMessages
    until not FQueryOpen;
    Result:= True;
  end;
end;

function TWaveOut.SetMaxCapsSettings: Boolean;
var aOpenFlags, aMaxSampling: LongWord;
begin
  Result:= False;
  if FIsOpen or not CheckDevice(FDeviceID) then
    Exit;
  ZeroMemory(@FCaps, SizeOf(FCaps));
  waveOutGetDevCaps(FDeviceID, @FCaps, SizeOf(FCaps));
  SetMaxFormatEx(FCaps.dwFormats, FFormatEx);
  FOpenFlags:= CALLBACK_WINDOW;
  if FCaps.dwSupport and WAVECAPS_SYNC = WAVECAPS_SYNC then
    FOpenFlags:= FOpenFlags or WAVE_ALLOWSYNC;
  if not CheckFormat then begin
    ErrMes('"' + DeviceName[FDeviceID] +
      '": Invalid wave format: ' + WaveFormat + '.');
    Exit;
  end;
  Result:= True;
  aOpenFlags:= FOpenFlags;
  with FFormatEx do begin
    if nSamplesPerSec = 44100 then begin
      nSamplesPerSec:= 48000;
      if not CheckFormat then
        nSamplesPerSec:= 44100
      else begin
        FOpenFlags:= aOpenFlags or WAVE_FORMAT_DIRECT;
        if not CheckFormat then
          FOpenFlags:= aOpenFlags;
        repeat
          aMaxSampling:= nSamplesPerSec;
          nSamplesPerSec:= aMaxSampling * 2;
          if nSamplesPerSec > cMaxSampling then begin
            aMaxSampling:= 48000;
            Break;
          end;
        until not CheckFormat;
        nSamplesPerSec:= aMaxSampling;
      end;
    end;
    InfMes('"' + DeviceName[FDeviceID] + '": Set format: ' + WaveFormat + '.');
  end;
end;

function TWaveOut.WaveFormat: String;
begin
  Result:= 'Channels=' + IntToStr(Channels) +
    ', Sampling=' + IntToStr(Sampling) + ', Bits=' + IntToStr(Bits);
end;

function TWaveOut.DevicesNumber: Integer;
begin
  Result:= waveOutGetNumDevs;
end;

function TWaveOut.GetDeviceName(aDevice: LongWord): String;
begin
  waveOutGetDevCaps(aDevice, @FCaps, SizeOf(FCaps));
  Result:= FCaps.szPname;
end;

procedure TWaveOut.SetChannels(aValue: Word);
begin
  if FIsOpen then
    Exit;
  FFormatEx.nChannels:= aValue;
end;

function TWaveOut.GetChannels: Word;
begin
  Result:= FFormatEx.nChannels;
end;

procedure TWaveOut.SetSampling(aValue: LongWord);
begin
  if FIsOpen then
    Exit;
  FFormatEx.nSamplesPerSec:= aValue;
end;

function TWaveOut.GetSampling: LongWord;
begin
  Result:= FFormatEx.nSamplesPerSec;
end;

procedure TWaveOut.SetBits(aValue: Word);
begin
  if FIsOpen then
    Exit;
  FFormatEx.wBitsPerSample:= aValue;
end;

function TWaveOut.GetBits: Word;
begin
  Result:= FFormatEx.wBitsPerSample;
end;

procedure TWaveOut.SetCount(aValue: LongWord);
begin
  if FIsOpen then
    Exit;
  FCount:= aValue;
end;

function TWaveOut.GetDataMatrix(aChannel, aIndex: Integer): Extended;
begin
  Result:= FDataMatrix[aChannel, aIndex];
end;

function TWaveOut.AllocBuffers: Boolean;
var i: Integer;
begin
  Result:= False;
  UpdateFormatEx(FFormatEx);
  SetLength(FWHdr, 2{Number of buffers});
  for i:= Low(FWHdr) to High(FWHdr) do with FWHdr[i] do begin
    ZeroMemory(@FWHdr[i], SizeOf(FWHdr[i]));
    dwBufferLength:= FCount * FFormatEx.nBlockAlign;
    lpData:= Pointer(GlobalAlloc(AllocFlags, dwBufferLength));
    if lpData = nil then begin
      ErrMes('Can not allocate buffer #' + IntToStr(i) + '.');
      Exit;
    end;
    dwLoops:= High(dwLoops);
    if i = Low(FWHdr) then
      dwFlags:= WHDR_BEGINLOOP;
    if i = High(FWHdr) then
      dwFlags:= dwFlags or WHDR_ENDLOOP;
  end;
  Result:= True;
end;

procedure TWaveOut.Open;
begin
  {$Q-}
  if (FDeviceID = High(FDeviceID)) or FIsOpen then
    Exit;
  UpdateFormatEx(FFormatEx);
  FQueryOpen:= False;
  FError:= WaveOutOpen(@FDeviceHD, FDeviceID, @FFormatEx, Handle, 0, FOpenFlags);
  if FError <> MMSYSERR_NOERROR then begin
    ErrMes('Open: ' + SysErrMes);
    Exit;
  end;
end;

procedure TWaveOut.OnWaveOpen(var Msg : TMessage);
var
  i, j, k, l, aBytes: Integer;
  aProc: procedure(var a; aValue: Extended);
begin
  if FQueryOpen then begin
    waveOutClose(FDeviceHD);
    Exit;
  end;
  FCloseQuery:= False;
  case Bits of
    8: aProc:= Proc8;
   16: aProc:= Proc16;
   else
     aProc:= Proc8;
  end;
  aBytes:= Bits shr 3;
  if not AllocBuffers then
    Exit;
  for i:= Low(FWHdr) to High(FWHdr) do with FWHdr[i] do begin
    l:= 0;
    try
      for j:= 0 to FCount - 1 do begin
        if l < High(FDataMatrix[0]) then
          Inc(l)
        else
          l:= 0;
        for k:= 0 to Channels - 1 do
          aProc(lpData[aBytes * (k + Channels * j)],  FDataMatrix[k,l]);
      end;
    except
      ErrMes('Error of copy data to header: Count=' +
          IntToStr(FCount) + ', Index=' + IntToStr(l));
      Exit;
    end;
    FError:= waveOutPrepareHeader(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Prepare header #' + IntToStr(i) + ': ' + SysErrMes);
      Exit;
    end;
    FError:= waveOutWrite(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Write header #' + IntToStr(i) + ': ' + SysErrMes);
      Exit;
    end;
    Inc(FLocked);
  end;
  FIsOpen:= True;
  if FIsOpen then begin
    if Assigned(FOnOpen) then
      FOnOpen(Self);
    InfMes('Start: ' + WaveFormat + '.');
  end;
end;

procedure TWaveOut.Trapezium(aSlopeTime, aFlatTime: Extended);
var
  i, j, aSlopeNum, aFlatNum: Integer;
begin
  if FIsOpen then
    Exit;
  try
    aSlopeNum:= Round(Max(1, aSlopeTime * Sampling));
    aFlatNum:= Round(aFlatTime * Sampling);
    FCount:= 2 * (aSlopeNum + aFlatNum);
    SetLength(FDataMatrix, Channels, FCount);
    j:= 0;
    for i:= 1 to aSlopeNum do begin
      FDataMatrix[0,j]:= -cAmp + (2 * cAmp * i) / aSlopeNum;
      Inc(j);
    end;
    for i:= 1 to aFlatNum do begin
      FDataMatrix[0,j]:= cAmp;
      Inc(j);
    end;
    for i:= 1 to aSlopeNum do begin
      FDataMatrix[0,j]:= cAmp - (2 * cAmp * i) / aSlopeNum;
      Inc(j);
    end;
    for i:= 1 to aFlatNum do begin
      FDataMatrix[0,j]:= - cAmp;
      Inc(j);
    end;
    for i:= 1 to High(FDataMatrix) do
      Move(FDataMatrix[0,0], FDataMatrix[i,0], FCount * SizeOf(Extended));
  except
    ErrMes('Exception in Trapezium.');
    Exit;
  end;
  Open;
end;

procedure TWaveOut.MultiSin(aFreqNum, aStep: Integer; aCount: LongWord);
var
  i, j: Integer;
  aAmp: Extended;
begin
  if FIsOpen then
    Exit;
  try
    FCount:= 1 shl 12;
    if Sampling > 150000 then
      FCount:= FCount shl 1
    else
      if Sampling < 75000 then
         FCount:= FCount shr 1;
    aAmp:= 0;
    SetLength(FDataMatrix, Channels, FCount);
    for i:= 0 to FCount - 1 do begin
      FDataMatrix[0,i]:= 0;
      for j:= 0 to aFreqNum do
        FDataMatrix[0,i]:= FDataMatrix[0,i]
                         + Sin(2*Pi * (1 shl (j*aStep))*i/(Pred(aCount)));
      if FDataMatrix[0,i] > aAmp then
        aAmp:= FDataMatrix[0,i];
    end;
    aAmp:= cAmp / aAmp;
    for i:= 0 to FCount - 1 do
      FDataMatrix[0,i]:= FDataMatrix[0,i] * aAmp;
    for i:= 1 to High(FDataMatrix) do
      Move(FDataMatrix[0,0], FDataMatrix[i,0], FCount * SizeOf(Extended));
  except
    ErrMes('Exception in MultiSin');
    Exit;
  end;
    Open;
end;

procedure TWaveOut.Close;
var i: Integer;
begin
  if FLocked > 0 then begin
    FCloseQuery:= True;
    FError:= waveOutReset(FDeviceHD);
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Close: Reset: ' + SysErrMes);
      Exit;
    end;
  end else if FIsOpen then begin
    for i:= Low(FWHdr) to High(FWHdr) do with FWHdr[i] do begin
      FError:= waveOutUnPrepareHeader(FDeviceHD, @FWHdr[i], SizeOf(FWHdr[i]));
      if FError <> MMSYSERR_NOERROR then begin
        ErrMes('Unprepare header #' + IntToStr(i) + ': ' + SysErrMes);
        Exit;
      end;
      lpData:= Pointer(GlobalFree(THandle(lpData)));
      if lpData <> nil then begin
        ErrMes('Can not free handle #' + IntToStr(i) + '.');
        Exit;
      end;
    end;
    FCloseQuery:= False;
    FError:= waveOutClose(FDeviceHD);
    if FError <> MMSYSERR_NOERROR then begin
      ErrMes('Close: ' + SysErrMes);
      Exit;
    end;
  end;
end;

procedure TWaveOut.OnWaveDone(var Msg : TMessage);
var
  WaveHeader: pWAVEHDR;
  hOutputDev: HWAVEOUT;
begin
  hOutputDev:= HWAVEOUT(Msg.WParam);
  if hOutputDev <> FDeviceHD then
    Exit;
  WaveHeader:= pWAVEHDR(Msg.lParam);
  if FCloseQuery then begin
    Dec(FLocked);
    if FLocked = 0 then begin
      FCloseQuery:= False;
      Close;
    end;
  end else
    waveOutWrite(hOutputDev, WaveHeader, SizeOf(WaveHeader^));
end;

procedure TWaveOut.OnWaveClose(var Msg : TMessage);
begin
  if FQueryOpen then begin
    FQueryOpen:= False;
    Exit;
  end;
  if FCloseQuery then
    Exit;
  FIsOpen:= False;
  InfMes('Stop.');
  if Assigned(FOnClose) then
    FOnClose(Self);
end;

function TWaveOut.SysErrMes: String;
var aErrorText: array[1..MAXERRORLENGTH] of Char;
begin
  if FError = MMSYSERR_NOERROR then
    Result:= ''
  else begin
    waveOutGetErrorText(FError, @aErrorText, SizeOf(aErrorText));
    Result:= aErrorText;
  end;
end;

end.
